#include "asmtrad.h"
#include "symtab_0.h"
#include "landef.h"
#include <stdio.h>
#include <stdlib.h>

static FILE *output;
static bool debug=0;
static typeid ident[MAX];

/* funzioni dello stack */
void varpop(typeid *s) {
	fprintf(output, "; varpop: \"%s\"\n", s);
	fprintf(output, "\tpop\teax\n");
	fprintf(output, "\tmov\t[%s], eax\n\n", s);
}

void varpush(typeid *s) {
	fprintf(output, "; varpush: \"%s\"\n", s);
	fprintf(output, "\tmov\teax, [%s]\n", s);
	fprintf(output, "\tpush\teax\n\n");
}

void valpush(value i) {
	fprintf(output, "; valpush\n");
	fprintf(output, "\tmov\teax, %d\n", i);
	fprintf(output, "\tpush\teax\n\n");
}

/* funzioni per le operazioni matematiche */
void addasm() {
	fprintf(output, "; somma\n");
	fprintf(output, "\tpop\tebx\n");
	fprintf(output, "\tpop\teax\n");
	fprintf(output, "\tadd\teax, ebx\n");
	if(debug)
		fprintf(output, "\tjo near\toverflow_label\t; se si verifica un overflow\n");
	fprintf(output, "\tpush\teax\n\n");
}

void subasm() {
	fprintf(output, "; sottrazione\n");
	fprintf(output, "\tpop\tebx\n");
	fprintf(output, "\tpop\teax\n");
	fprintf(output, "\tsub\teax, ebx\n");
	if(debug)
		fprintf(output, "\tjo near\toverflow_label\n");
	fprintf(output, "\tpush\teax\n\n");
}

void mulasm() {
	fprintf(output, "; moltiplicazione\n");
	fprintf(output, "\tpop\tebx\n");
	fprintf(output, "\tpop\teax\n");
	fprintf(output, "\timul\tebx\n");
	if(debug)
		fprintf(output, "\tjo near\toverflow_label\n");
	fprintf(output, "\tpush\teax\n\n");
}

void divasm() {
	fprintf(output, "; divisione\n");
	fprintf(output, "\tpop\tebx\n");
	if(debug) {
		fprintf(output, "\tcmp\tebx, 0\n");
		fprintf(output, "\tje near\tdivide_label\n");
	}
	fprintf(output, "\tpop\teax\n");
	fprintf(output, "\tcdq\n");
	fprintf(output, "\tidiv\tebx\n");
	fprintf(output, "\tpush\teax\n\n");
}

void modasm() {
	fprintf(output, "; modulo\n");
	fprintf(output, "\tpop\tebx\n");
	if(debug) {
		fprintf(output, "\tcmp\tebx, 0\n");
		fprintf(output, "\tje near\tdivide_label\n");
	}
	fprintf(output, "\tpop\teax\n");
	fprintf(output, "\tcdq\n");
	fprintf(output, "\tidiv\tebx\n");
	fprintf(output, "\tpush\tedx\n\n");
}

/* funzioni prologo e epilogo */
void prologo() {
	fprintf(output, "; --- INIZIO FILE ASSMBLER ---\n\n");

	if(debug)
		fprintf(output, "; debug attivato\n\n");
	else
		fprintf(output, "; debug disattivato\n\n");

	fprintf(output, "; data: dati inizializzati\n");
	fprintf(output, "segment .data\n");
	fprintf(output, "\tint_format\tdb\t\"%%d\",0\t; terminatore NULL del C\n");
	fprintf(output, "\tstring_format\tdb\t\"%%s\",0\n");
	fprintf(output, "\theader\t\tdb\t\"-- Valore finale delle Variabili --\",0\n");
	fprintf(output, "\tvariabile\tdb\t\"Variabile: \",0\n");
	fprintf(output, "\tvalore\t\tdb\t\"Valore: \",0\n");
	fprintf(output, "\tnotused\t\tdb\t\"Non utilizzata\",0\n");

	if(debug) {
		fprintf(output, "\toverflow_msg\tdb\t\"** Error: int overflow!\",0\n");
		fprintf(output, "\tdivide_zero\tdb\t\"** Error: divide by zero!\",0\n");
	}

	tabgetinit();
	while(tabgetnext(ident))
			fprintf(output, "\tvar_%s\t\tdb\t\"%s\",0\n", ident, ident);
	fprintf(output, "\n");

	fprintf(output, "; bss: dati non inizializzati\n");
	fprintf(output, "segment .bss\n");
	tabgetinit();
	while(tabgetnext(ident))
			fprintf(output, "\t%s\t\tresd\t1\n", ident);
	fprintf(output, "\n");

	fprintf(output, "; text: codice\n");
	fprintf(output, "section .text\n");
	fprintf(output, "\tglobal asm_main\t\t; esporto la routine asm_main\n");
	fprintf(output, "\textern printf, putchar\t; funzioni definite altrove (lib c)\n\n");

	fprintf(output, "asm_main:\n");
	fprintf(output, "\tenter 0,0\t\t; routine di setup\n");
	fprintf(output, "\tpusha\t\t\t; push di tutti i registri\n\n");

	fprintf(output, "; ------ fine prologo ------\n\n");
}

void epilogo() {
	fprintf(output, "; ------ inizio epilogo ------\n\n");

	fprintf(output, "; stampo l'header\n");
	fprintf(output, "\tmov eax, header\t\t; metto in eax l'indirizzo della stringa\n");
	fprintf(output, "\tcall print_string\t; chiamo routine stampa\n");
	fprintf(output, "\tcall print_nl\t\t; stampa un carattere new line\n\n");

	tabgetinit();
	while(tabgetnext(ident)) {
		fprintf(output, "; stampo la variabile %s e il suo valore\n", ident);
		fprintf(output, "\tmov eax, variabile\n");
		fprintf(output, "\tcall print_string\n");
		fprintf(output, "\tmov eax, var_%s\n", ident);
		fprintf(output, "\tcall print_string\n");
		fprintf(output, "\tcall print_tab\n");

		if(tabcheckinit(ident)) {
			fprintf(output, "\tmov eax, valore\n");
			fprintf(output, "\tcall print_string\n");
			fprintf(output, "\tmov eax, [%s]\n", ident);
			fprintf(output, "\tcall print_int\n");
			fprintf(output, "\tcall print_nl\n");
		}
		else {
			fprintf(output, "\tmov eax, notused\n");
			fprintf(output, "\tcall print_string\n");
			fprintf(output, "\tcall print_nl\n");

		}
	}

	fprintf(output, "\tpopa\n");
	fprintf(output, "\tmov eax, 0\t\t; ritorna al codice C\n");
	fprintf(output, "\tleave\n");
	fprintf(output, "\tret\n\n");

	fprintf(output, "; stampa un intero\n");
	fprintf(output, "print_int:\n");
	fprintf(output, "\tenter 0,0\n");
	fprintf(output, "\tpusha\n");
	fprintf(output, "\tpushf\n\n");

	fprintf(output, "\tpush eax\n");
	fprintf(output, "\tpush dword int_format\n");
	fprintf(output, "\tcall printf\n");
	fprintf(output, "\tpop ecx\n");
	fprintf(output, "\tpop ecx\n\n");

	fprintf(output, "\tpopf\n");
	fprintf(output, "\tpopa\n");
	fprintf(output, "\tleave\n");
	fprintf(output, "\tret\n\n");

	fprintf(output, "; stampa una stringa\n");
	fprintf(output, "print_string:\n");
	fprintf(output, "\tenter 0,0\n");
	fprintf(output, "\tpusha\n");
	fprintf(output, "\tpushf\n\n");

	fprintf(output, "\tpush eax\n");
	fprintf(output, "\tpush dword string_format\n");
	fprintf(output, "\tcall printf\n");
	fprintf(output, "\tpop ecx\n");
	fprintf(output, "\tpop ecx\n\n");

	fprintf(output, "\tpopf\n");
	fprintf(output, "\tpopa\n");
	fprintf(output, "\tleave\n");
	fprintf(output, "\tret\n\n");

	fprintf(output, "; stampa un carattere di newline\n");
	fprintf(output, "print_nl:\n");
	fprintf(output, "\tenter 0,0\n");
	fprintf(output, "\tpusha\n");
	fprintf(output, "\tpushf\n\n");

	fprintf(output, "\tpush eax\n");
	fprintf(output, "\tpush dword 10\t\t; 10 == ASCII \\n\n");
	fprintf(output, "\tcall putchar\n");
	fprintf(output, "\tpop ecx\n");
	fprintf(output, "\tpop ecx\n\n");

	fprintf(output, "\tpopf\n");
	fprintf(output, "\tpopa\n");
	fprintf(output, "\tleave\n");
	fprintf(output, "\tret\n\n");

	fprintf(output, "; stampa un carattere di tabulazione\n");
	fprintf(output, "print_tab:\n");
	fprintf(output, "\tenter 0,0\n");
	fprintf(output, "\tpusha\n");
	fprintf(output, "\tpushf\n\n");

	fprintf(output, "\tpush eax\n");
	fprintf(output, "\tpush dword 9\t\t; 9 == ASCII \\t\n");
	fprintf(output, "\tcall putchar\n");
	fprintf(output, "\tpop ecx\n");
	fprintf(output, "\tpop ecx\n\n");

	fprintf(output, "\tpopf\n");
	fprintf(output, "\tpopa\n");
	fprintf(output, "\tleave\n");
	fprintf(output, "\tret\n\n");

	if(debug) {
		fprintf(output, "; stampa un errore e torna al chimante\n");
		fprintf(output, "overflow_label:\n");
		fprintf(output, "\tmov eax, overflow_msg\t; stampa l'errore\n");
		fprintf(output, "\tcall print_string\n");
		fprintf(output, "\tcall print_nl\t\t; stampa un carattere new line\n\n");
		fprintf(output, "\tpopa\n");
		fprintf(output, "\tmov eax, 1\t\t; ritorna al codice C\n\n");
		fprintf(output, "\tleave\n");
		fprintf(output, "\tret\n\n");

		fprintf(output, "; stampa un errore e torna al chimante\n");
		fprintf(output, "divide_label:\n");
		fprintf(output, "\tmov eax, divide_zero\t; stampa l'errore\n");
		fprintf(output, "\tcall print_string\n");
		fprintf(output, "\tcall print_nl\t\t; stampa un carattere new line\n\n");
		fprintf(output, "\tpopa\n");
		fprintf(output, "\tmov eax, 1\t\t; ritorna al codice C\n\n");
		fprintf(output, "\tleave\n");
		fprintf(output, "\tret\n\n");
	}

	fprintf(output, "; --- FINE FILE ASSMBLER ---\n");
}

void inittrad(char *file_out) {
	if((output=fopen(file_out, "w"))==NULL) {
		puts("** Error: apertura file destinazione");
		exit (1);
	}
}

void closetrad() {
	fclose(output);
}

void enable_debug(bool enable) {
	debug=enable;
}
