#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "parser_1.h"
#include "landef.h"
#include "device.h"
#include "num_conv.h"
#include "asmtrad.h"

/* define utili per le chiamate all'assembler e al linker */
#define NASM "nasm -f elf "
#define DST_FILE "program"
#define ASM ".asm"
#define LINKER_1 "gcc driver.o "
#define LINKER_2 " -o "

/* chiama le funzioni che sostituiscono il nodo 
dell'albero semantico con il corrispondente assembler */
static void tradeexportorder(tree t) {
	switch(t->tag) {
		case CST: valpush(t->U.val); break;
		case VRB: varpush(t->U.stringaid); break;
		case OPR: 
			tradeexportorder(t->sin);
			tradeexportorder(t->des);
			switch(t->U.opcode) {
				case PLUS: addasm(); break;
				case MINUS: subasm(); break;
				case TIMES: mulasm(); break;
				case DIVSY: divasm(); break;
				case MODSY: modasm(); break;
				default: break;
			}
	}
}

/* se il nodo  un becomes */
static void passo2statement(tree t) {
	tradeexportorder(right(t));
	varpop((left(t))->U.stringaid);
}

/* analisi albero semantico */
static void passo2(tree t) {
	if (t->tag==OPR) {
		if (t->U.opcode==SEMICOLON) {
			passo2(left(t));
			passo2(right(t));
		}
		else
			passo2statement(t);
	}
}

int main(int argc, char *argv[]) {
	int num_base=BASE;
	char src[MAX_FILE]=SOURCE_FILE;
	char dst[MAX_FILE]=DST_FILE;
	tree albero_sematico;
	char *file_asm;

	/* variabili per i parametri della riga di comando */
	int i=1;
	char minus, flag;

	/* variabili per la creazione dell'eseguibile */
	char *nasm_str, *linker_str;
	bool eseguibile=true;

	/* gestione parametri riga di comando */
	if(argc==1)
		puts("(Per altre opzioni chiama \"./compiler -h\")");
	else {
		while(i<argc) {
			sscanf(argv[i++], "%c%c", &minus, &flag);
			if(minus=='-') {
				switch(flag) {
					case 'b':
					sscanf(argv[i++], "%d", &num_base);
					printf("Settata base a %d\n", num_base);
					break;

					case 'c':
					eseguibile=false;
					puts("Creer solo il file asm");
					break;

					case 'g':
					enable_debug(true);
					puts("Codice di debug attivato");
					break;

					case 'h':
					puts("Usage: ./compiler [options]");
					puts("Options:");
					puts("-b <arg>\tSetta la base. Default 10");
					puts("-c\t\tCrea solo il file asm");
					puts("-g\t\tAggiunge controlli runtime in assembler. Default false");
					puts("-h\t\tVisualizza questo menu");
					printf("-i <file>\tSpecifica file di input. Default %s\n", SOURCE_FILE);
					printf("-o <file>\tSpecifica file di output. Default %s\n", DST_FILE);
					puts("-V\t\tVisualizza la versione e alcune informazioni");
					exit (0);
					break;

					case 'i':
					strcpy(src, argv[i++]);
					printf("File di input impostato a: %s\n", src);
					break;

					case 'o':
					strcpy(dst, argv[i++]);
					printf("File di output impostato a: %s\n", dst);
					break;

					case 'V':
					puts("Mp compiler versione 0.0.0.0.1");
					puts("Unife - 2006/2007");
					puts("CdL Informatica - Linguaggi 1");
					puts("Andrea Galvani - 068172");
					exit (0);
					break;

					default:
					puts("** Error: parametro passato non riconosciuto");
					exit (0);
					break;
				}
			}
			else {
				puts("** Error: parametro passato non riconosciuto");
				exit (1);
			}
		}
	}

	/* apro il file di input controllo */
	if(initdevice(src)) {
		puts("** Error: apertura file sorgente");
		exit (1);
	}

	/* setto la base */
	setbase(num_base);

	/* chiamo il parser che crea l'albero */
	albero_sematico=program();

	if(albero_sematico==NULL) {
		puts("** Error: l'albero  vuoto, c' qualcosa che tragia nel sorgente");
		exit (1);
	}

	/* file_asm = "dst" + ".asm" */
	file_asm=(char *)calloc(strlen(dst)+strlen(ASM), sizeof(char));
	strcat(file_asm, dst);
	strcat(file_asm, ASM);

	/* apro il file di output */
	inittrad(file_asm);

	/* inizio la scrittura nel file assembly */
	prologo();

	/* compilazione della forma intermedia */
	passo2(albero_sematico);

	/* finisco la scrittura nel file assembly */
	epilogo();

	/* chiudo file output */
	closetrad();

	/* chiudo il file di input */
	closedevice();

	printf("Generazione del file %s completata ;)\n", dst);

	/* automatizzo la creazione dell'eseguibile dal file .asm */
	if(eseguibile) {
		/* stringa per chimare nasm: nasm -f elf dst.asm */
		nasm_str=(char *)calloc(strlen(NASM)+strlen(file_asm), sizeof(char));
		strcat(nasm_str, NASM);
		strcat(nasm_str, file_asm);

		/* chiamo nasm */
		if(system(nasm_str))
			puts("** Error: nasm ha trovato qualcosa che non va :(");
		else {
			puts("File oggetto creato correttamente ;)");

			/* stringa per chimara linker: gcc dst.o driver.o -o dst */
			linker_str=(char *)calloc(strlen(LINKER_1)+strlen(dst)+strlen(".o")+strlen(LINKER_2)+strlen(dst), sizeof(char));
			strcat(linker_str, LINKER_1);	/* gcc driver.o */
			strcat(linker_str, dst);
			strcat(linker_str, ".o");	/* dst.o */
			strcat(linker_str, LINKER_2);	/* -o */
			strcat(linker_str, dst);	/* dst */

			/* chiamo il linker */
			if(system(linker_str))
				puts("** Error: gcc non  riuscito a linkare :(");
			else
				printf("Eseguibile \"%s\" creato correttamente ;)\n", dst);
		}
	}

	return 0;
}
