#include <stdio.h>
#include "parser_0.h"
#include "lex.h"
#include "landef.h"
#include "symtab_0.h"

static bool fact();
static bool term();
static bool expression();
static bool statement();
static bool listofstat();
static bool listofid();
static bool vardecl();
bool program();

/* <FACT> -> IDENT | INTCON | LPAR<EXPRESSION>RPAR */
static bool fact() {
	switch(sy) {
		case IDENT:
			if(tabexist(id))
				if(tabcheckinit(id)) {
					insymbol();
					return true;
				}
				else {
					printf("Parse error (fact): variabile %s non inizializzata\n", id);
					return false;
				}
			else {
				printf("Parse error (fact): variabile %s non dichiarata\n", id);
				return false;
			}

		case INTCON:
			insymbol();
			return true;

		case LPAR:
			insymbol();
			if(expression()) {
				if(sy==RPAR) {
					insymbol();
					return true;
				}
				else {
					puts("Parse error (fact): attesa una \")\"");
					return false;
				}
			}
			else {
				puts("Parse error (fact): attesa un'espressione");
				return false;
			}

		default:
			puts("Parse error (fact): simbolo non riconosciuto");
			return false;
	}
}

/* <TERM> -> <FACT> {[TIMES|DIVSY|MODSY] <FACT>} */
static bool term() {
	if(fact()) {
		while(sy==TIMES || sy==DIVSY || sy==MODSY) {
			insymbol();
			if(!fact()) {
				puts("Parse error (term): atteso un fattore (2)");
				return false;
			}
		}
		return true;
	}
	else {
		puts("Parse error (term): atteso un fattore (1)");
		return false;
	}
}

/* <EXPRESSION> -> <TERM> {[PLUS|MINUS] <TERM>} */
static bool expression() {
	if(term()) {
		while(sy==PLUS || sy==MINUS) {
			insymbol();
			if(!term()) {
				puts("Parse error (expression): atteso un term (2)");
				return false;
			}
		}
		return true;
	}
	else {
		puts("Parse error (expression): atteso un term (1)");
		return false;
	}
}

/* <STATEMENT> -> IDENT BECOMES <EXPRESSION> */
static bool statement() {
	if(sy==IDENT) {
		if(tabexist(id)) {
			tabsetinit(id);
			insymbol();
			if(sy==BECOMES) {
				insymbol();
				if(expression())
					return true;
				else {
					puts("Parse error (statement): atteso un espressione");
					return false;
				}
			}
			else {
				puts("Parse error (statement): atteso \":=\"");
				return false;
			}
		}
		else {
			printf("Parse error (statement): Variabile \"%s\" non dichiarata\n", id);
			return false;
		}
	}
	else {
		puts("Parse error (statement): atteso un identificatore");
		return false;
	}
}

/* <LISTOFSTAT> -> <STATEMENT> {SEMICOLON <STATEMENT>} */
static bool listofstat() {
	if(statement()) {
		while(sy==SEMICOLON) {
			insymbol();
			if(!statement()) {
				puts("Parse error (listofstat): atteso uno statement");
				return false;
			}
		}
		return true;
	}
	else {
		puts("Parse error (listofstat): atteso uno statement");
		return false;
	}
}

/* <LISTOFID> -> IDENT {COLON IDENT} */
static bool listofid() {
	if(sy==IDENT) {
		tabinsert(id);
		insymbol();
		while(sy==COLON) {
			insymbol();
			if(sy==IDENT) {
				if(tabexist(id))
					printf("Parse error (listofid): identificatore %s doppio", id);
				else
					tabinsert(id);
			}
			else {
				puts("Parse error (listofid): atteso un identificatore");
				return false;
			}
			insymbol();
		}
		return true;
	}
	else {
		puts("Parse error (listofid): atteso un identificatore");
		return false;
	}
}

/* <VARDECL> -> VARSY <LISTOFID> SEMICOLON */
static bool vardecl() {
	if(sy==VARSY) {
		insymbol();
		if(listofid()) {
			if(sy==SEMICOLON) {
				insymbol();
				return true;
			}
			else {
				puts("Parse error (vardecl): atteso \";\"");
				return false;
			}
		}
		else {
			puts("Parse error (vardecl): atteso un identificatore");
			return false;
		}
	}
	else {
		puts("Parse error (vardecl): attesa variabile");
		return false;
	}
}

/* <PROGRAM> -> <VARDECL> <LISTOFSTAT> FULLSTOP */
bool program() {
	initlex();

	if(vardecl()) {
		if(listofstat()) {
			if(sy==FULLSTOP) {
				return true;
			}
			else {
				puts("Parse error (program): atteso \".\"");
				return false;
			}
		}
		else {
			puts("Parse error (program): atteso uno statement");
			return false;	
		}
	}
	else {
		puts("Parse error (program): attesa una dichiarazione di variabile");
		return false;
	}
}
