#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "parser_1.h"
#include "lex.h"
#include "tabs.h"
#include "landef.h"
#include "bintree.h"
#include "symtab_0.h"

static tree evalop(symbol sy, tree t1, tree t2);
static tree eval(symbol sy);
static tree fact();
static tree term();
static tree expression();
static tree statement();
static tree listofstat();
static bool listofid();
static bool vardecl();

/* <PROGRAM> -> <VARDECL> <LISTOFSTAT> FULLSTOP */
tree program() {
	initlex();

	tree t1=NULL;

	if(vardecl()) {
		if((t1=listofstat())!=NULL) {
			if(sy==FULLSTOP) {
				return t1;
			}
			else {
				puts("Parse error (program): atteso \".\"");
				return t1;
			}
		}
		else {
			puts("Parse error (program): atteso uno statement");
			return t1;	
		}
	}
	else {
		puts("Parse error (program): attesa una dichiarazione di variabile");
		return t1;
	}
}

/* <VARDECL> -> VARSY <LISTOFID> SEMICOLON */
static bool vardecl() {
	if(sy==VARSY) {
		insymbol();
		if(listofid()) {
			if(sy==SEMICOLON) {
				insymbol();
				return true;
			}
			else {
				puts("Parse error (vardecl): atteso \";\"");
				return false;
			}
		}
		else {
			puts("Parse error (vardecl): atteso un identificatore");
			return false;
		}
	}
	else {
		puts("Parse error (vardecl): attesa variabile");
		return false;
	}
}

/* <LISTOFID> -> IDENT {COLON IDENT} */
static bool listofid() {
	if(sy==IDENT) {
		tabinsert(id);
		insymbol();
		while(sy==COLON) {
			insymbol();
			if(sy==IDENT) {
				if(tabexist(id))
					printf("Parse error (listofid): identificatore %s doppio", id);
				else
					tabinsert(id);
			}
			else {
				puts("Parse error (listofid): atteso un identificatore");
				return false;
			}
			insymbol();
		}
		return true;
	}
	else {
		puts("Parse error (listofid): atteso un identificatore");
		return false;
	}
}

/* <LISTOFSTAT> -> <STATEMENT> {SEMICOLON <STATEMENT>} */
static tree listofstat() {
	tree t1=NULL, t2=NULL;

	if((t1=statement())!=NULL) {
		while(sy==SEMICOLON) {
			insymbol();
			if((t2=statement())!=NULL) {
				t1=evalop(SEMICOLON, t1, t2);
			}
			else {
				puts("Parse error (listofstat): atteso uno statement");
				return t1;
			}
		}
		return t1;
	}
	else {
		puts("Parse error (listofstat): atteso uno statement");
		return t1;
	}
}

/* <STATEMENT> -> IDENT BECOMES <EXPRESSION> */
static tree statement() {
	typeid leftid[MAX];
	tree t1=NULL, rightval;

	if(sy==IDENT) {
		if(!tabexist(id)) {
			printf("Parse error (statement): variabile \"%s\" non dichiarata\n", id);
			return t1;
		}
		else {
			tabsetinit(id);
			strcpy(leftid, id);
			insymbol();

			if(sy==BECOMES) {
				insymbol();

				t1=(tree)malloc(sizeof(struct treenode));
				t1->tag=VRB;
				t1->des=NULL;
				t1->sin=NULL;
				strcpy(t1->U.stringaid, leftid);

				if((rightval=expression())!=NULL)
					return evalop(BECOMES, t1, rightval);
				else {
					puts("Parse error (statement): attesa un espressione");
					return t1;
				}
			}
			else {
				puts("Parse error (statement): atteso \":=\"");
				return t1;
			}
		}
	}
	else {
		puts("Parse error (statement): atteso un identificatore");
		return t1;
	}
}

/* <EXPRESSION> -> <TERM> {[PLUS|MINUS] <TERM>} */
static tree expression() {
	symbol op;
	tree t1=NULL, t2=NULL;

	if((t1=term())==NULL) {
		puts("Parse error (expression): atteso un term");
		return t1;
	}
	else {
		while(sy==PLUS || sy==MINUS) {
			op=sy;
			insymbol();
			if((t2=term())==NULL) {
				puts("Parse error (expression): atteso un term");
				return t1;
			}
			else
				t1=evalop(op, t1, t2);
		}
		return t1;
	}
}

/* <TERM> -> <FACT> {[TIMES|DIVSY|MODSY] <FACT>} */
static tree term() {
	symbol op;
	tree t1=NULL, t2=NULL;

	if((t1=fact())==NULL) {
		puts("Parse error (term): atteso un fattore");
		return t1;
	}
	else {
		while(sy==TIMES || sy==DIVSY || sy==MODSY) {
			op=sy;
			insymbol();
			if((t2=fact())==NULL) {
				puts("Parse error (term): atteso un fattore");
				return t1;
			}
			else
				t1=evalop(op, t1, t2);
		}
		return t1;
	}
}

/* <FACT> -> IDENT | INTCON | LPAR<EXPRESSION>RPAR */
static tree fact() {
	tree t1=NULL;

	switch(sy) {
		case IDENT:
			if(tabexist(id)) {
				if(tabcheckinit(id)) {
					t1=eval(sy);
					insymbol();
					return t1;
				}
				else {
					printf("Parse error (fact): variabile \"%s\" non inizializzata\n", id);
					return t1;
				}
			}
			else {
				printf("Parse error (fact): variabile \"%s\" non dichiarata\n", id);
				return t1;
			}

		case INTCON:
			t1=eval(sy);
			insymbol();
			return t1;

		case LPAR:
			insymbol();
			if((t1=expression())!=NULL) {
				if(sy!=RPAR) {
					puts("Parse error (fact): attesa una \")\"");
					exit(1);
				}
				else {
					insymbol();
					return t1;
				}
			}
			else {
				puts("Parse error (fact): attesa un'espressione");
				return t1;
			}

		default:
			puts("Parse error (fact): simbolo non riconosciuto");
			return t1;
	}
}

/* costruisce una foglia dell'abero semantico */
static tree eval(symbol sy) {
	tree temp=constree(NULL, NULL);


	if(sy==INTCON) {
		temp->tag=CST;
		temp->U.val=inum;
	}

	if(sy==IDENT) {
		temp->tag=VRB;
		strcpy(temp->U.stringaid, id);
	}

	return temp;
}

/* costruisce un nodo dell'albero semantico */
static tree evalop(symbol sy, tree t1, tree t2) {
	tree temp=constree(t1, t2);

	temp->tag=OPR;
	temp->U.opcode=sy;

	return temp;
}
