#include "symtab_0.h"
#include "landef.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

/* struttura symtab */
struct symtab {
	typeid name[MAX];
	value val;
	bool init;
	struct symtab *sin;
	struct symtab *des;
};

/* struttura lista */
struct lista {
	typeid name[MAX];
	struct lista *next;
};

typedef struct symtab *node_ptr;
typedef struct lista *list_ptr;

static node_ptr radice=NULL;
static list_ptr testa=NULL;

/* alloca una foglia */
static node_ptr addnode(typeid *el) {
	node_ptr node;

	node=(node_ptr)malloc(sizeof(struct symtab));

	/* setta il nome dell'identificatore */
	strcpy(node->name, el);
	/* inizializzo per comodita val a zero */
	node->val=0;
	/* setta il flag inizializzato a false */
	node->init=false;

	node->sin=NULL;
	node->des=NULL;

	return node;
}

/* posiziona la nuova foglia al suo posto */
static void nonemptyinsert(typeid *el) {
	node_ptr new_node=addnode(el);
	node_ptr node=radice;
	bool inserito=false;

	while(!inserito) {
		if(strcmp(el, node->name)<0) {
			/* visita il figlio sinistro */
			if(node->sin==NULL) {
				node->sin=new_node;
				inserito=true;
			}
			else
				node=node->sin;
		}
		else {
			/* visita il figlio destro */
			if(node->des==NULL) {
				node->des=new_node;
				inserito=true;
			}
			else
				node=node->des;
		}
	}
}

void tabinsert(typeid *el) {
	if(radice==NULL)
		radice=addnode(el);
	else
		nonemptyinsert(el);
}

bool tabexist(typeid *el) {
	node_ptr node=radice;
	int i;

	while(1) {
		i=strcmp(el, node->name);

		if(i==0)
			return true;
		else if(i<0) {
			if(node->sin==NULL)
				return false;
			else
				node=node->sin;
		}
		else {
			if(node->des==NULL)
				return false;
			else
				node=node->des;
		}
	}
}

void tabsetinit(typeid *el) {
	node_ptr node=radice;
	int i;
	bool settato=false;

	while(!settato) {
		i=strcmp(el, node->name);

		if(i==0) {
			node->init=true;
			settato=true;
		}
		else if(i<0) {
			if(node->sin!=NULL)
				node=node->sin;
		}
		else {
			if(node->des!=NULL)
				node=node->des;
		}
	}
}

bool tabcheckinit(typeid *el) {
	node_ptr node=radice;
	int i;

	while(1) {
		i=strcmp(el, node->name);

		if(i==0)
			return node->init;
		else if(i<0) {
			if(node->sin!=NULL)
				node=node->sin;
		}
		else {
			if(node->des!=NULL)
				node=node->des;
		}
	}
}

void tabsetvalue(typeid *el, value valore) {
	node_ptr node=radice;
	int i;
	bool settato=false;

	while(!settato) {
		i=strcmp(el, node->name);

		if(i==0) {
			node->val=valore;
			settato=true;
		}
		else if(i<0) {
			if(node->sin!=NULL)
				node=node->sin;
		}
		else {
			if(node->des!=NULL)
				node=node->des;
		}
	}
}

value tabreadvalue(typeid * el) {
	node_ptr node=radice;
	int i;

	while(1) {
		i=strcmp(el, node->name);

		if(i==0)
			return node->val;
		else if(i<0) {
			if(node->sin!=NULL)
				node=node->sin;
		}
		else {
			if(node->des!=NULL)
				node=node->des;
		}
	}
}

/* alloca una cella della lista */
static void addcella(typeid *el) {
	list_ptr list;

	list=(list_ptr)malloc(sizeof(struct lista));

	/* setta il nome dell'identificatore */
	strcpy(list->name, el);

	list->next=testa;
	testa=list;
}

/* crea la lista con gli identificatori dell'albero */
static void crealista(node_ptr node) {
	if(node!=NULL) {
		if(node->sin==NULL && node->des==NULL) {
			addcella(node->name);
		}
		else {
			crealista(node->des);
			addcella(node->name);
			crealista(node->sin);
		}
	}
}

/* restituisce il primo elemento dalla lista */
void tabgetinit() {
	crealista(radice);
}

/* restituisce un elemento dalla lista */
bool tabgetnext(typeid *name) {
	list_ptr list=testa;

	if(testa==NULL)
		return false;
	else {
		strcpy(name, testa->name);
		testa=testa->next;
		free(list);
		return true;
	}
}

/* stampa il contenuto del nodo */
static void printnode(node_ptr node) {
	printf("%s-%d-%d\n", node->name, node->val, node->init);
}

/* stampa la struttura dell'albero */
static void printtab2(int displace, node_ptr node) {
	int i;

	if(node!=NULL) {
		/* se  una foglia */
		if(node->sin==NULL && node->des==NULL) {
			for(i=0; i<displace; i++)
				printf("\t");
			printnode(node);
		}
		else {
			printtab2(displace+1, node->des);
			for(i=0; i<displace; i++)
				printf("\t");
			printnode(node);
			printtab2(displace+1, node->sin);
		}
	}
}

void printtab() {
	node_ptr node=radice;
	
	if(node==NULL)
		puts("** Error: non stamper niente. La symtab  vuota!");
	else
		printtab2(0, node);
}
