#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bintree.h"
#include "parser_1.h"
#include "landef.h"
#include "limits.h"
#include "device.h"
#include "num_conv.h"
#include "symtab_0.h"

/* funzione che restituisce il valore associato il nodo */
static value postorder(tree t) {
	value t1, t2;
	int long temp;

	switch(t->tag) {
		case CST: return t->U.val;
		case VRB: return tabreadvalue(t->U.stringaid);
		case OPR:
			t1=postorder(left(t));
			t2=postorder(right(t));
			switch(t->U.opcode) {
				case PLUS:
					if((temp=t1+t2)>INT_MAX) {
						puts("** Error: int overflow");
						exit(1);
					}
					else
						return temp;
					
				case MINUS:
					if((temp=t1-t2)<INT_MIN) {
						puts("** Error: int underflow");
						exit(1);
					}
					else
						return temp;

				case TIMES:
					if((temp=t1*t2)>INT_MAX) {
						puts("** Error: int overflow");
						exit(1);
					}
					else
						return temp;

				case DIVSY:
					if(t2==0) {
						puts("** Error: divide by zero");
						exit(1);
					}
					else
						return t1/t2;

				case MODSY:
					if(t2==0) {
						puts("** Error: divide by zero");
						exit(1);
					}
					else
						return t1%t2;

				default: break;
			}
	}

	return 0;
}

/* se il nodo  un becomes */
static void passo2statement(tree t) {
	value val=postorder(right(t));
	tabsetvalue((left(t))->U.stringaid, val);
}

/* analisi albero semantico */
static void passo2(tree t) {
	if (t->tag==OPR) {
		if (t->U.opcode==SEMICOLON) {
			passo2(left(t));
			passo2(right(t));
		}
		else
			passo2statement(t);
	}
}

int main(int argc, char *argv[]) {
	int num_base=BASE, i=1;
	char src[MAX_FILE]=SOURCE_FILE;
	char minus, flag;
	bool file_error;
	typeid ident[MAX];
	tree albero_sematico;

	/* gestione parametri riga di comando */
	if(argc==1)
		puts("(Per altre opzioni chiama \"./t_interpreter -h\")");
	else {
		while(i<argc) {
			sscanf(argv[i++], "%c%c", &minus, &flag);
			if(minus=='-') {
				switch(flag) {
					case 'b':
					sscanf(argv[i++], "%d", &num_base);
					printf("Settata base a %d\n", num_base);
					break;

					case 'h':
					puts("Usage: ./t_interpreter [options]");
					puts("Options:");
					puts("-b <arg>\tSetta la base. Default 10");
					puts("-h\t\tVisualizza questo menu");
					printf("-i <file>\tSpecifica file di input. Default %s\n", SOURCE_FILE);
					puts("-V\t\tVisualizza la versione e alcune informazioni");
					exit (0);
					break;

					case 'i':
					strcpy(src, argv[i++]);
					printf("File di input impostato a: %s\n", src);
					break;

					case 'V':
					puts("Mp t_interpreter versione 0.0.0.0.1");
					puts("Unife - 2006/2007");
					puts("CdL Informatica - Linguaggi 1");
					puts("Andrea Galvani - 068172");
					exit (0);
					break;

					default:
					puts("** Error: parametro passato non riconosciuto");
					exit (0);
					break;
				}
			}
			else {
				puts("** Error: parametro passato non riconosciuto");
				exit (1);
			}
		}
	}

	/* apro il file di input controllo */
	if((file_error=initdevice(src))==true) {
		puts("** Error: apertura file sorgente");
		exit (1);
	}

	/* setto la base */
	setbase(num_base);

	/* interpretazione della forma intermedia */
	albero_sematico=program();
	if(albero_sematico==NULL)
		puts("** Error: l'albero  vuoto, c' qualcosa che tragia nel sorgente");
	else
		passo2(albero_sematico);

	/* chiudo il file di input */
	closedevice();

	/* stampo il valore delle variabili */
	puts("-- Valore finale delle Variabili --");
	tabgetinit();
	while(tabgetnext(ident)) {
		if(tabcheckinit(ident)==0)
			printf("Variabile: %s\tNon utilizzata\n", ident);
		else
			printf("Variabile: %s\tValore: %d\n", ident, tabreadvalue(ident));
	}

	return 0;
}
